from astropy.nddata import CCDData
import astropy.stats as stats
from astropy.visualization.wcsaxes import WCSAxes
from madcubapy.io import MadcubaMap
import matplotlib as mpl
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

__all__ = [
    'add_wcs_axes',
    'add_manual_wcs_axes',
    'add_colorbar',
    'insert_colorbar',
    'parse_clabel',
]


def add_wcs_axes(
        fig=None,
        nrows=1,
        ncols=1,
        number=1,
        fitsmap=None,
        use_std=False,
        **kwargs):
    """
    Add a `~astropy.visualization.wcsaxes.WCSAxes` object with WCS
    coordinates into an existing figure.

    A `~matplotlib.figure.Figure` with no axes has to be set before calling
    this function. This is due to the inability to change axis coordinates
    after it has been called. The coordinates have to be called when creating
    the axes object. The function returns objects for the axes and the image.

    Parameters
    ----------
    fig : `~matplotlib.figure.Figure`
        Figure to which the WCSAxes are added.
    nrows : `~int`
        Number of rows on the subplot grid.
    ncols : `~int`
        Number of columns on the subplot grid.
    number : `~int`
        Number of subplot in the grid in which to paint the Axes.
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map to be displayed.
    use_std : `~bool`
        If true, set color limits to ±3 times the standard deviation of the
        image data.

    Returns
    -------
    ax : `~astropy.visualization.wcsaxes.WCSAxes`
        Axes object with the selected map coordinates.
    img : `~matplotlib.image.AxesImage`
        Image object of the selected map.

    Other Parameters
    ----------------
    **kwargs
        Parameters to pass to :func:`matplotlib.pyplot.imshow`.

    """
    
    if not fig:
        raise TypeError(
            "add_wcs_axes() missing 1 required positional argument: 'fig'"
        )
    elif not isinstance(fig, mpl.figure.Figure):
        raise TypeError(
            f"'fig' argument must be a {mpl.figure.Figure}"
        )
    if fitsmap == None:
        raise TypeError(
            f"Need to specify a fitsmap parameter.")
    elif (not isinstance(fitsmap, MadcubaMap) and
          not isinstance(fitsmap, CCDData)):
        raise TypeError(
            f"'fitsmap' argument must be a {MadcubaMap} or {CCDData}")
    if fitsmap.wcs is None:
        raise TypeError(
            f"This fits file has problems in its WCS parameters.")

    data = fitsmap.data
    wcs = fitsmap.wcs.celestial
    # store BUNIT in a global variable
    global last_bunit
    last_bunit = parse_clabel(fitsmap)

    # Slice extra dimensions from data
    if fitsmap.header['NAXIS'] == 3:
        data = data[0, :, :]
    elif fitsmap.header['NAXIS'] == 4:
        data = data[0, 0, :, :]
    # Dimensions are: stokes, freq, Y, X for ALMA datacubes

    if use_std:
        mean, median, std = stats.sigma_clipped_stats(data, sigma=3.0)
        kwargs['vmin'] = median - 3 * std
        kwargs['vmax'] = median + 3 * std

    # Create the WCS axes
    ax = fig.add_subplot(nrows, ncols, number, projection=wcs)
    img = ax.imshow(data, **kwargs)
    ax.coords[0].set_axislabel("RA (ICRS)")
    ax.coords[1].set_axislabel("DEC (ICRS)")
    # Vertical ticklabel for DEC
    ax.coords[1].set_ticklabel(rotation='vertical', rotation_mode='default')
    # Set title for MadcubaMaps
    if isinstance(fitsmap, MadcubaMap):
        ax.set_title(fitsmap.filename)

    return ax, img


def add_manual_wcs_axes(
        fig=None,
        left=0, 
        bottom=0, 
        width=1, 
        height=1,
        fitsmap=None,
        use_std=False,
        **kwargs):
    """
    Add a `~astropy.visualization.wcsaxes.WCSAxes` object with WCS
    coordinates in a manually set position into an existing figure.

    A `~matplotlib.figure.Figure` with no axes has to be set before calling
    this function. This is due to the inability to change axis coordinates
    after it has been called. The coordinates have to be called when creating
    the axes object. The function returns objects for the axes and the image.

    Parameters
    ----------
    fig : `~matplotlib.figure.Figure`
        Figure to which the WCSAxes is added.
    left : `~float`
        X coordinate to begin the Axes subplot.
    bottom : `~float`
        Y coordinate to begin the Axes subplot.
    width : `~float`
        Width of the Axes subplot.
    height : `~float`
        Height of the Axes subplot.
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map to be displayed.
    use_std : `~bool`
        If true, set color limits to ±3 times the standard deviation of the
        image data.

    Returns
    -------
    ax : `~astropy.visualization.wcsaxes.WCSAxes`
        Axes object with the selected map coordinates.
    img : `~matplotlib.image.AxesImage`
        Image object of the selected map.

    Other Parameters
    ----------------
    **kwargs
        Parameters to pass to :func:`matplotlib.pyplot.imshow`.

    """
    
    if not fig:
        raise TypeError(
            "add_manual_wcs_axes() missing 1 required positional argument: 'fig'"
        )
    elif not isinstance(fig, mpl.figure.Figure):
        raise TypeError(
            f"'fig' argument must be a {mpl.figure.Figure}"
        )
    if fitsmap == None:
        raise TypeError(
            f"Need to specify a fitsmap parameter.")
    elif (not isinstance(fitsmap, MadcubaMap) and
          not isinstance(fitsmap, CCDData)):
        raise TypeError(
            f"'fitsmap' argument must be a {MadcubaMap} or {CCDData}")
    if fitsmap.wcs is None:
        raise TypeError(
            f"This fits file has problems in its WCS parameters.")
    
    data = fitsmap.data
    wcs = fitsmap.wcs.celestial
    # store BUNIT in a global variable
    global last_bunit
    last_bunit = parse_clabel(fitsmap)

    # Slice extra dimensions from data
    if fitsmap.header['NAXIS'] == 3:
        data = data[0, :, :]
    elif fitsmap.header['NAXIS'] == 4:
        data = data[0, 0, :, :]
    # Dimensions are: stokes, freq, Y, X for ALMA datacubes

    if use_std:
        mean, median, std = stats.sigma_clipped_stats(data, sigma=3.0)
        kwargs['vmin'] = median - 3 * std
        kwargs['vmax'] = median + 3 * std

    # Create the WCS axes
    ax = fig.add_axes([left, bottom, width, height], projection=wcs)
    img = ax.imshow(data, **kwargs)
    ax.coords[0].set_axislabel("RA (ICRS)")
    ax.coords[1].set_axislabel("DEC (ICRS)")
    # Vertical ticklabel for DEC
    ax.coords[1].set_ticklabel(rotation='vertical', rotation_mode='default')
    # Set title for MadcubaMaps
    if isinstance(fitsmap, MadcubaMap):
        ax.set_title(fitsmap.filename)

    return ax, img


def parse_clabel(fitsmap):
    """
    Parse colorbar text from a `~madcubapy.io.MadcubaMap` or
    `~astropy.nddata.CCDData` unit attribute.

    Parameters
    ----------
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map object from which to extract units information.

    Returns
    -------
    label : `~str`
        Label to be used in the colorbar.

    """

    units = fitsmap.unit.to_string()
    if units == 'Jy / beam':
        label = r'$I \ {\rm (Jy \ beam^{-1})}$'
    elif units == 'Jy m / (beam s)':
        label = r'$I \ {\rm (Jy \ beam^{-1} \ m \ s^{-1})}$'
    elif units == 'km mJy / (beam s)':
        label = r'$I \ {\rm (mJy \ beam^{-1} \ km \ s^{-1})}$'
    elif units == 'Jy km / (beam s)':
        label = r'$I \ {\rm (Jy \ beam^{-1} \ km \ s^{-1})}$'
    elif units == 'm mJy / (beam s)':
        label = r'$I \ {\rm (mJy \ beam^{-1} \ m \ s^{-1})}$'
    else:
        label = 'units not parsed'

    return label


def add_colorbar(
        ax=None,
        location='right',
        size=0.05,
        pad=0.03,
        **kwargs):
    """
    Add a colorbar to one side of a `~matplotlib.axes.Axes` or
    `~astropy.visualization.wcsaxes.WCSAxes` object.

    Parameters
    ----------
    ax : `~matplotlib.axes.Axes` or `~astropy.visualization.wcsaxes.WCSAxes`
        Axes to which the colorbar is added.
    location : {"left", "right", "bottom", "top"}
        Where the colorbar is positioned relative to the main
        `~astropy.visualization.wcsaxes.WCSAxes`.
    size : `~float`
        Fraction of 'ax' to use as size for the colorbar.
    pad : `~float`
        Separation between the colorbar bar and 'ax'.

    Returns
    -------
    cbar : `~matplotlib.colorbar.Colorbar`
        Colorbar object.

    Other Parameters
    ----------------
    **kwargs
        Parameters to pass to :func:`matplotlib.pyplot.colorbar`.

    """

    if ax == None:
        raise TypeError(
            f"Need to specify an axes object with 'ax' keyword.")

    # Get figure and image objects from the axes
    fig = ax.get_figure()
    img = ax.get_images()[0]

    # Add colorbar
    if location == 'left' or location == 'right':
        orientation = 'vertical'
    elif location == 'top' or location == 'bottom':
        orientation = 'horizontal'
    else:
        raise ValueError(
            f"location can only be 'top', 'right', 'bottom', or 'left'")
    # Use bunit from last fitsmap plotted if present
    if 'label' not in kwargs:
        try:
            last_bunit
        except NameError:
            kwargs['label'] = 'units not found'
        else:
            kwargs['label'] = last_bunit
    ax_xini, ax_yini, ax_width, ax_height = ax.get_position(
        original=False).bounds

    if location == 'right':
        cax = fig.add_axes([ax_xini+ax_width+pad*(ax_width),
                            ax_yini,
                            size*ax_width,
                            ax_height])
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='vertical', **kwargs)
        colorbar.ax.tick_params(
            axis="y",
            which="both",
            left=False,
            right=True,
            labelleft=False,
            labelright=True,
        )
        colorbar.ax.yaxis.set_label_position('right')
        if isinstance(ax, WCSAxes):
            ax.coords[1].tick_params(
                axis="y",
                which="both",
                labelleft=True,
                labelright=False,
            )
            ax.coords[1].set_axislabel_position('l')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labelleft=True,
                labelright=False,
            )
    elif location == 'left':
        cax = fig.add_axes([ax_xini-pad*(ax_width)-size*(ax_width),
                            ax_yini,
                            size*ax_width,
                            ax_height])
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='vertical', **kwargs)
        colorbar.ax.tick_params(
            axis="y",
            which="both",
            left=True,
            right=False,
            labelleft=True,
            labelright=False,
        )
        colorbar.ax.yaxis.set_label_position('left')
        if isinstance(ax, WCSAxes):
            ax.coords[1].tick_params(
                axis="y",
                which="both",
                labelleft=False,
                labelright=True,
            )
            ax.coords[1].set_axislabel_position('r')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labelleft=False,
                labelright=True,
            )
    elif location == 'top':
        cax = fig.add_axes([ax_xini,
                            ax_yini+ax_height+pad*(ax_height),
                            ax_width,
                            size*ax_height])
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='horizontal', **kwargs)
        colorbar.ax.tick_params(
            axis="x",
            which="both",
            top=True,
            bottom=False,
            labeltop=True,
            labelbottom=False,
        )
        colorbar.ax.xaxis.set_label_position('top')
        if isinstance(ax, WCSAxes):
            ax.coords[0].tick_params(
                axis="x",
                which="both",
                labeltop=False,
                labelbottom=True,
            )
            ax.coords[0].set_axislabel_position('b')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labeltop=False,
                labelbottom=True,
            )
        ax.set_title(None)
    elif location == 'bottom':
        cax = fig.add_axes([ax_xini,
                            ax_yini-pad*(ax_height)-size*(ax_height),
                            ax_width,
                            size*ax_height])
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='horizontal', **kwargs)
        colorbar.ax.tick_params(
            axis="x",
            which="both",
            top=False,
            bottom=True,
            labeltop=False,
            labelbottom=True,
        )
        colorbar.ax.xaxis.set_label_position('bottom')
        if isinstance(ax, WCSAxes):
            ax.coords[0].tick_params(
                axis="x",
                which="both",
                labeltop=True,
                labelbottom=False
            )
            ax.coords[0].set_axislabel_position('t')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labeltop=True,
                labelbottom=False,
            )
        ax.set_title(None)

    return colorbar


def insert_colorbar(
        ax=None,
        location='right',
        size='5%',
        pad=0.05,
        **kwargs):
    """
    Insert a colorbar to one side of a `~matplotlib.axes.Axes` or 
    `~astropy.visualization.wcsaxes.WCSAxes` object, fitting it into the same
    space.

    Parameters
    ----------
    ax : `~matplotlib.axes.Axes` or `~astropy.visualization.wcsaxes.WCSAxes`
        Axes to which the colorbar is added.
    location : {"left", "right", "bottom", "top"}
        Where the colorbar is positioned relative to the main
        `~astropy.visualization.wcsaxes.WCSAxes`.
    size : `~str`
        Size percentage of 'ax' to use as size for the colorbar.
    pad : `~float`
        Separation between the colorbar bar and 'ax'.

    Returns
    -------
    cbar : `~matplotlib.colorbar.Colorbar`
        Colorbar object.

    Other Parameters
    ----------------
    **kwargs
        Parameters to pass to :func:`matplotlib.pyplot.colorbar`.

    """

    if ax == None:
        raise TypeError(
            f"Need to specify an axes object with 'ax' keyword.")

    # Get figure and image objects from the axes
    fig = ax.get_figure()
    img = ax.get_images()[0]

    # Add colorbar
    if location == 'left' or location == 'right':
        orientation = 'vertical'
    elif location == 'top' or location == 'bottom':
        orientation = 'horizontal'
    else:
        raise ValueError(
            f"location can only be 'top', 'right', 'bottom', or 'left'")
    # Use bunit from last fitsmap plotted if present
    if 'label' not in kwargs:
        try:
            last_bunit
        except NameError:
            kwargs['label'] = 'units not found'
        else:
            kwargs['label'] = last_bunit
    divider = make_axes_locatable(ax)
    cax = divider.append_axes(location, size=size, 
                              pad=pad, axes_class=Axes)
    if location == 'right':
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='vertical', **kwargs)
        colorbar.ax.tick_params(
            axis="y",
            which="both",
            left=False,
            right=True,
            labelleft=False,
            labelright=True,
        )
        colorbar.ax.yaxis.set_label_position('right')
        if isinstance(ax, WCSAxes):
            ax.coords[1].tick_params(
                axis="y",
                which="both",
                labelleft=True,
                labelright=False,
            )
            ax.coords[1].set_axislabel_position('l')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labelleft=True,
                labelright=False,
            )
    elif location == 'left':
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='vertical', **kwargs)
        colorbar.ax.tick_params(
            axis="y",
            which="both",
            left=True,
            right=False,
            labelleft=True,
            labelright=False,
        )
        colorbar.ax.yaxis.set_label_position('left')
        if isinstance(ax, WCSAxes):
            ax.coords[1].tick_params(
                axis="y",
                which="both",
                labelleft=False,
                labelright=True,
            )
            ax.coords[1].set_axislabel_position('r')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labelleft=False,
                labelright=True,
            )
    elif location == 'top':
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='horizontal', **kwargs)
        colorbar.ax.tick_params(
            axis="x",
            which="both",
            top=True,
            bottom=False,
            labeltop=True,
            labelbottom=False,
        )
        colorbar.ax.xaxis.set_label_position('top')
        if isinstance(ax, WCSAxes):
            ax.coords[0].tick_params(
                axis="x",
                which="both",
                labeltop=False,
                labelbottom=True,
            )
            ax.coords[0].set_axislabel_position('b')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labeltop=False,
                labelbottom=True,
            )
        ax.set_title(None)
    elif location == 'bottom':
        colorbar = fig.colorbar(img, cax=cax,
                                orientation='horizontal', **kwargs)
        colorbar.ax.tick_params(
            axis="x",
            which="both",
            top=False,
            bottom=True,
            labeltop=False,
            labelbottom=True,
        )
        colorbar.ax.xaxis.set_label_position('bottom')
        if isinstance(ax, WCSAxes):
            ax.coords[0].tick_params(
                axis="x",
                which="both",
                labeltop=True,
                labelbottom=False
            )
            ax.coords[0].set_axislabel_position('t')
        else:
            ax.tick_params(
                axis="both",
                which="both",
                top=True,
                bottom=True,
                left=True,
                right=True,
                labeltop=True,
                labelbottom=False,
            )
        ax.set_title(None)

    return colorbar



from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs import WCS
from madcubapy.io import MadcubaMap
from madcubapy.visualization import add_wcs_axes
from madcubapy.visualization import copy_zoom_axes
from madcubapy.visualization import copy_zoom_fitsmap
from madcubapy.visualization import sync_zoom
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
import numpy as np
import pytest

@pytest.fixture
def example_madcuba_map_tm1():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "examples/data/IRAS16293_SO2a_tm1.fits"
    )

@pytest.fixture
def example_madcuba_map_tm2():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "examples/data/IRAS16293_SO2a_tm2.fits"
    )


@pytest.mark.parametrize("invalid_input", [
    None, 123, "string", [], {}, set(), object(), plt.figure().add_subplot(111)
])
def test_sync_zoom_incorrent_types(invalid_input):
    with pytest.raises(TypeError):
        sync_zoom(invalid_input)


def test_limits_results(example_madcuba_map_tm1, example_madcuba_map_tm2):
    # Fitsmap
    xlim1 = (200, 350)
    ylim1 = (200, 350)
    xlim_fitsmap, ylim_fitsmap = copy_zoom_fitsmap(
        ref_fitsmap=example_madcuba_map_tm1,
        target_fitsmap=example_madcuba_map_tm2,
        x_lim=(xlim1), 
        y_lim=(ylim1)
    )
    # Axes
    fig1 = plt.figure(figsize=(3,2))
    ax1, img1 = add_wcs_axes(fig1, 1, 2, 1, fitsmap=example_madcuba_map_tm1)
    ax2, img2 = add_wcs_axes(fig1, 1, 2, 2, fitsmap=example_madcuba_map_tm2)
    # Copy zoom
    ax1.set_xlim(xlim1)
    ax1.set_ylim(ylim1)
    copy_zoom_axes(ref_ax=ax1, target_ax=ax2)
    xlim_axes = ax2.get_xlim()
    ylim_axes = ax2.get_ylim()
    # Sync zoom
    fig2 = plt.figure(figsize=(3,2))
    ax3, img3 = add_wcs_axes(fig2, 1, 3, 1, fitsmap=example_madcuba_map_tm1)
    ax4, img4 = add_wcs_axes(fig2, 1, 3, 2, fitsmap=example_madcuba_map_tm2)
    # Sync zoom
    sync_zoom(ax3, ax4)
    ax3.set_xlim(xlim1)
    xlim_sync = ax4.get_xlim()
    ylim_sync = ax4.get_ylim()
    
    np.testing.assert_allclose(xlim_fitsmap, xlim_axes)
    np.testing.assert_allclose(xlim_fitsmap, xlim_sync)



from astropy.visualization.wcsaxes import WCSAxes
import numpy as np
from madcubapy.coordinates import transform_coords_axes
from madcubapy.coordinates import transform_coords_fitsmap

__all__ = [
    'copy_zoom_fitsmap',
    'copy_zoom_axes',
    'sync_zoom',
]


def copy_zoom_fitsmap(
        ref_fitsmap,
        target_fitsmap,
        x_lim,
        y_lim,
        origin=0):
    """Copy a map's limits to another map.

    Parameters
    ----------
    ref_fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Reference mape.
    target_fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map into which limits are transformed.
    x_lim : array-like
        X axis limits from ref_fitsmap.
    y_lim : array-like
        Y axis limits from ref_fitsmap.
    origin : int
        Origin of the coordinates of the image. 0 for numpy standards,
        and 1 for FITS standards.

    Returns
    -------
    new_x_lim : `list`
        Transformed X axis limits.
    new_y_lim : `list`
        Transformed Y axis limits.

    """

    if len(x_lim) != 2 or len(y_lim) != 2:
        raise TypeError(f"Limits need to be delimited by two coordinates " 
                        + "each (min, max)")

    # Image corners
    bottom_left = np.array([x_lim[0], y_lim[0]])
    top_right = np.array([x_lim[1], y_lim[1]])
    # Transform limits
    new_bottom_left = transform_coords_fitsmap(ref_fitsmap=ref_fitsmap,
                                               target_fitsmap=target_fitsmap,
                                               points=bottom_left,
                                               origin=origin)
    new_top_right = transform_coords_fitsmap(ref_fitsmap=ref_fitsmap,
                                             target_fitsmap=target_fitsmap,
                                             points=top_right,
                                             origin=origin)
    # Return new limits
    new_x_lim = [new_bottom_left[0], new_top_right[0]]
    new_y_lim = [new_bottom_left[1], new_top_right[1]]

    return new_x_lim, new_y_lim


def copy_zoom_axes(
        ref_ax,
        target_ax):
    """Copy a `~astropy.visualization.wcsaxes.WCSAxes` limits to another
    `~astropy.visualization.wcsaxes.WCSAxes`.

    Parameters
    ----------
    ref_ax : `~astropy.visualization.wcsaxes.WCSAxes`
        Reference Axes.
    target_ax : `~astropy.visualization.wcsaxes.WCSAxes`
        Axes into which limits are transformed.

    """

    # Read limits
    x_lim = ref_ax.get_xlim()
    y_lim = ref_ax.get_ylim()
    # Image corners
    bottom_left = np.array([x_lim[0], y_lim[0]])
    top_right = np.array([x_lim[1], y_lim[1]])
    # Transform limits
    new_bottom_left = transform_coords_axes(ref_ax=ref_ax,
                                            target_ax=target_ax,
                                            points=bottom_left)
    new_top_right = transform_coords_axes(ref_ax=ref_ax,
                                          target_ax=target_ax,
                                          points=top_right)
    # Set new limits
    target_ax.set_xlim(new_bottom_left[0], new_top_right[0])
    target_ax.set_ylim(new_bottom_left[1], new_top_right[1])


def sync_zoom(
        *axes):
    """Synchronize X and Y limits between any number of
    `~astropy.visualization.wcsaxes.WCSAxes` objects.

    Parameters
    ----------
    *axes : `~astropy.visualization.wcsaxes.WCSAxes`
        WCSAxes objects to synchronize.

    Notes
    _____
    Do not use this function too many times. It seems that python stores
    every axes object connected in a given jupyter kernel run. If the
    session runs for too long, the program will slow down considerably.

    """

    # Check that all provided axes are instances of WCSAxes
    if not all(isinstance(ax, WCSAxes) for ax in axes):
        raise TypeError("All arguments must be instances of WCSAxes.")

    class LimitSyncer:
        def __init__(self):
            # Shared state to prevent recursion
            self.updating = False

        def sync_x_limits(self, ax):
            if self.updating:
                return
            self.updating = True
            try:
                # Read limits from reference axes
                x_lim = ax.get_xlim()
                y_lim = ax.get_ylim()
                # Image corners
                bottom_left = np.array([x_lim[0], y_lim[0]])
                top_right = np.array([x_lim[1], y_lim[1]])
                for axis in axes:
                    if axis != ax:
                        # Transform limits
                        new_bottom_left = transform_coords_axes(
                            ref_ax=ax,
                            target_ax=axis,
                            points=bottom_left
                        )
                        new_top_right = transform_coords_axes(
                            ref_ax=ax,
                            target_ax=axis,
                            points=top_right
                        )
                        # Set new limits in target axes
                        axis.set_xlim(new_bottom_left[0], new_top_right[0])
            finally:
                self.updating = False

        def sync_y_limits(self, ax):
            if self.updating:
                return
            self.updating = True
            try:
                # Read limits from reference axes
                x_lim = ax.get_xlim()
                y_lim = ax.get_ylim()
                # Image corners
                bottom_left = np.array([x_lim[0], y_lim[0]])
                top_right = np.array([x_lim[1], y_lim[1]])
                for axis in axes:
                    if axis != ax:
                        # Transform limits
                        new_bottom_left = transform_coords_axes(
                            ref_ax=ax,
                            target_ax=axis,
                            points=bottom_left
                        )
                        new_top_right = transform_coords_axes(
                            ref_ax=ax,
                            target_ax=axis,
                            points=top_right
                        )
                        # Set new limits in target axes
                        axis.set_ylim(new_bottom_left[1], new_top_right[1])
            finally:
                self.updating = False

    # Create an instance of the limit syncer
    syncer = LimitSyncer()

    # Connect the limit change events to the syncer methods
    # For this loop, it is needed to pass ax to the lambda function
    for ax in axes:
        ax.callbacks.connect('xlim_changed',
                             lambda event, ax=ax : syncer.sync_x_limits(ax))
        ax.callbacks.connect('ylim_changed',
                             lambda event, ax=ax : syncer.sync_y_limits(ax))



# madcubapy/coordinates/__init__.py

"""
The `madcubapy.coordinates` package provides functions to help the user
work with coordinates and transformations.
"""

from .offsets import *
from .transforms import *



import astropy
from astropy.io import fits
from astropy.nddata import CCDData
from astropy.table import Table
import astropy.units as u
from copy import deepcopy
import numpy as np
import os
from pathlib import Path

from .madcubafits import MadcubaFits

__all__ = [
    'MadcubaMap',
]

class MadcubaMap(MadcubaFits):
    """
    A container for MADCUBA FITS maps, using the `~madcubapy.io.MadcubaFits`
    interface.

    This class is basically a wrapper to read MADCUBA exported FITS maps and
    their history files with astropy.

    Parameters
    ----------
    data : `~numpy.ndarray`
        The data array contained in the FITS file.
    header : `~astropy.io.fits.Header`
        The header object associated with the FITS file.
    wcs : `~astropy.wcs.WCS`
        Object with the world coordinate system for the data in the FITS file.
    unit : `~astropy.units.Unit`
        The unit of the data in the FITS file.
    hist : `~astropy.table.Table`
        Table containing the history information of the FITS file, which is
        stored in a separate *_hist.csv* file.
    ccddata : `~astropy.nddata.CCDData`
        An astropy CCDData object loaded with astropy as a failsafe.
    filename : `~str`
        Name of the FITS file.

    Methods
    -------
    add_hist(*args)
        Load the history table from a csv file.

    """
    
    def __init__(
        self,
        data=None,
        header=None,
        wcs=None,
        unit=None,
        hist=None,
        ccddata=None,
        filename=None,
    ):
        # inherit hist
        super().__init__(hist)  # Initialize the parent class with hist

        if data is not None and not isinstance(data, np.ndarray):
            raise TypeError("The data must be a numpy array.")
        self._data = data

        if header is not None and not isinstance(header, astropy.io.fits.Header):
            raise TypeError("The header must be an astropy.io.fits.Header.")
        self._header = header

        if wcs is not None and not isinstance(wcs, astropy.wcs.WCS):
            raise TypeError("The WCS must be an astropy.wcs.WCS.")
        self._wcs = wcs

        if unit is not None and not isinstance(unit, astropy.units.UnitBase):
            raise TypeError("The unit must be an astropy unit.")
        self._unit = unit

        if ccddata is not None and not isinstance(ccddata, astropy.nddata.CCDData):
            raise TypeError("The ccddata must be a CCDData instance.")
        self._ccddata = ccddata

        if filename is not None and not isinstance(filename, str):
            raise TypeError("The filename must be a string.")
        self._filename = filename

    @property
    def ccddata(self):
        """
        `~astropy.nddata.CCDData` : An astropy CCDData object loaded with
        astropy as a failsafe.
        """
        return self._ccddata

    @ccddata.setter
    def ccddata(self, value):
        if value is not None and not isinstance(value, CCDData):
            raise TypeError("The ccddata must be a CCDData instance.")
        self._ccddata = value

    @property
    def filename(self):
        """
        `~str` : Name of the FITS file.
        """
        return self._filename

    @filename.setter
    def filename(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("The filename must be a string.")
        self._filename = value

    @property
    def data(self):
        """
        `~numpy.ndarray` : The data array contained in the FITS file.
        """
        return self._data

    @data.setter
    def data(self, value):
        if value is not None and not isinstance(value, np.ndarray):
            raise TypeError("The data must be a numpy array.")
        self._data = value

    @property
    def header(self):
        """
        `~astropy.io.fits.Header` : The header object associated with the FITS
        file.
        """
        return self._header

    @header.setter
    def header(self, value):
        if value is not None and not isinstance(value, astropy.io.fits.Header):
            raise TypeError("The header must be an astropy.io.fits.Header.")
        self._header = value

    @property
    def wcs(self):
        """
        `~astropy.wcs.WCS` : Object with the world coordinate system for the data in the FITS file.
        """
        return self._wcs

    @wcs.setter
    def wcs(self, value):
        if value is not None and not isinstance(value, astropy.wcs.WCS):
            raise TypeError("The WCS must be an astropy.wcs.WCS.")
        self._wcs = value

    @property
    def unit(self):
        """
        `~astropy.units.Unit` : The unit of the data in the FITS file.
        """
        return self._unit

    @unit.setter
    def unit(self, value):
        if value is not None and not isinstance(value, astropy.units.UnitBase):
            raise TypeError("The unit must be an astropy unit.")
        self._unit = value

    @classmethod
    def read(cls, filepath, **kwargs):
        """
        ``Classmethod`` to generate a `~madcubapy.io.MadcubaMap` object from a
        FITS file. This method creates an `~astropy.nddata.CCDData` from the
        FITS file.

        Parameters
        ----------
        filepath : `~str`
            Name of FITS file.

        Other Parameters
        ----------------
        **kwargs
            Additional keyword parameters passed through to the Astropy
            :func:`~astropy.nddata.fits_ccddata_reader` function.

        """
        fits_filepath = filepath
        filename_terms = str(filepath).split('/')
        filename = filename_terms[-1]
        # Check if the fits file exists
        if not os.path.isfile(fits_filepath):
            raise FileNotFoundError(f"File {fits_filepath} not found.")
        # Load the CCDData from the .fits file
        ccddata = CCDData.read(fits_filepath, **kwargs)
        # Load the Table from the .csv file if present
        hist_filepath = os.path.splitext(fits_filepath)[0] + "_hist.csv"
        if not os.path.isfile(hist_filepath):
            print("WARNING: Default history file not found.")
            hist = None
        else:
            hist = Table.read(hist_filepath, format='csv')
        # Store the attributes
        data = ccddata.data
        header = ccddata.header
        # header = fits.getheader(fits_filepath)
        wcs = ccddata.wcs
        unit = ccddata.unit
        # Return an instance of MadcubaFits
        madcuba_map = cls(
            data=data,
            header=header,
            wcs=wcs,
            unit=unit,
            hist=hist,
            ccddata=ccddata,
            filename=filename,
        )
        if madcuba_map.hist:
            update_action = f"Open cube: '{str(filepath)}'"
            madcuba_map._update_hist(update_action)
        return madcuba_map

    def write(self, filepath, **kwargs):
        """
        Write a `~madcubapy.io.MadcubaMap` into a FITS file alongside its
        history file.

        Parameters
        ----------
        filepath : `~str`
            Name of output FITS file.
        
        Other Parameters
        ----------------
        **kwargs
            Additional keyword parameters passed through to the Astropy
            :func:`~astropy.nddata.fits_ccddata_writer` function.

        """
        if not self._ccddata:
            raise TypeError("Cannot export manually created MadcubaMaps (yet)")
        else:
            # Get save directory and file name
            filepath_terms = str(filepath).split('/')
            save_dir = Path("/".join(filepath_terms[:-1]))
            filename = filepath_terms[-1]
            filename_terms = filename.split('.')
            csv_filename = ".".join(filename_terms[:-1]) + "_hist.csv"
            # Write fits
            self._ccddata.write(filepath, **kwargs)
            # Correct units not recognized by madcuba
            with fits.open(filepath, mode="update") as hdul:
                parsed_bunit = _parse_madcuba_friendly_bunit(self.unit)
                hdul[0].header['BUNIT'] = parsed_bunit
                hdul.flush()  # Save changes
            # write hist
            if self.hist:
                update_action = f"Save cube: '{str(filepath)}'"
                self._update_hist(update_action)
                if 'overwrite' in kwargs:
                    overwrite_csv = kwargs['overwrite']
                else:
                    overwrite_csv = False
                self.hist.write(
                    save_dir/csv_filename,
                    format='csv',
                    overwrite=overwrite_csv,
                ) 
            else:
                print("Empty history file has not been saved")
            # Update filename
            self.filename = filename

    def copy(self):
        """
        Create a copy of the `~madcubapy.io.MadcubaMap`.
        """
        if self._hist:
            new_hist = self._hist.copy()
        else:
            new_hist = None
        return MadcubaMap(
            data=deepcopy(self._data),
            header=deepcopy(self._header),
            wcs=deepcopy(self._wcs),
            unit=deepcopy(self._unit),
            hist=new_hist,
            ccddata=deepcopy(self._ccddata),
        )

    def show(self, **kwargs):
        """
        Show the map in a pop-up window.

        Other Parameters
        ----------------
        **kwargs
            Additional parameters passed to
            :func:`~madcubapy.visualization.add_wcs_axes`.
            
        """
        from madcubapy.visualization.quick_plotters import quick_show
        quick_show(self, **kwargs)

    def get_input(self, **kwargs):
        """
        Return mouse click coordinates from this map.

        Other Parameters
        ----------------
        **kwargs
            Additional parameters passed to
            :func:`~madcubapy.visualization.add_wcs_axes`.
        """
        from madcubapy.visualization.interaction import _get_input_from_map
        return _get_input_from_map(self, **kwargs)

    def update_sigma(self, statistic='std', **kwargs):
        """
        Measure the noise (sigma) of the map by calculating the standard
        deviation (std) or root mean square (rms) inside several polygons
        selected by mouse clicks, and store it in the SIGMA header card.

        - Left clicks create polygon vertices.
        - Right click closes the current polygon, and a subsequent left click
          starts a new polygon.

        Parameters
        ----------
        statistic : {'std', 'rms'}, optional
            Statistic to be used as sigma. Defaults to 'std' and can be changed
            at runtime via GUI buttons.

        Other Parameters
        ----------------
        **kwargs
            Additional parameters passed to
            :func:`~madcubapy.visualization.add_wcs_axes`.
        
        """
        from madcubapy.operations.maps.noise import measure_noise
        sigma = measure_noise(self, statistic, **kwargs)
        # Update sigma header card
        if np.isnan(sigma):
            raise Exception("Function aborted.")
        self.header["SIGMA"] = (sigma, 'madcubapy update sigma')
        self.ccddata.header["SIGMA"] = (sigma, 'madcubapy update sigma')
        # Update hist file
        if self._hist:
            self._update_hist((f"Update sigma to '{sigma}'."))

    def fix_units(self):
        """
        Tries to fix problems when the units are incorrectly parsed. The user
        should confirm that the new units are correct.
        """
        unit_str = self.header["BUNIT"]
        # Fix CARTA strings
        new_unit_str = _fix_unit_string_multiple_slashes(unit_str)
        # Overwrite units
        self._unit = u.Unit(new_unit_str)
        self._ccddata.unit = u.Unit(new_unit_str)
        if self._hist:
            self._update_hist((f"Fixed BUNIT card from '{unit_str}' "
                             + f"to '{new_unit_str}"))

    def convert_unit_to(self, unit):
        """
        Convert the units of the map to other units.
        """
        previous_unit = self.unit
        # Change unit in CCDDdata and copy it into MadcubaMap
        converted_ccddata = self._ccddata.convert_unit_to(unit)
        self._ccddata = converted_ccddata
        self._data = converted_ccddata.data
        self._unit = converted_ccddata.unit
        # Convert sigma header card
        if "SIGMA" in self.header:
            sigma = self.header["SIGMA"] * previous_unit
            converted_sigma = sigma.to(unit).value
            self.header["SIGMA"] = (converted_sigma, 'madcubapy convert unit')
            self.ccddata.header["SIGMA"] = (converted_sigma,
                                            'madcubapy convert unit')
        if self._hist:
            self._update_hist((f"Convert units to "
                             + f"'{unit.to_string(format='fits')}'"))

    def __repr__(self):
        # If hist is None, display that it's missing
        if self._hist is None:
            hist_r = "hist=None"
        # If hist is present, display a summary of the table
        else: hist_r = (
            f"hist=<Table length={len(self._hist)} rows, " +
            f"{len(self._hist.columns)} columns>"
        )
        if self._data is None:
            data_r = "data=None"
        else:
            data_r = f"data=<numpy.ndarray shape={self._data.shape}>"
        if self._unit is None:
            unit_r = "unit=None"
        else:
            unit_r = f"unit={self._unit}"

        return f"<MadcubaMap({data_r}, {unit_r}, {hist_r})>"



def _fix_unit_string_multiple_slashes(unit_str):
    """
    This function converts dots to spaces and slashes to '-1' exponents if the
    BUNIT card contains more than one slash.
    """
    result = []
    # Split by slashes
    terms = unit_str.split('/')
    # The entire first term is in the numerator, no correction for a slash must
    # be applied. Split the units and append to a list.
    first_sub_terms = terms[0].split('.')
    result.extend(first_sub_terms)
    # Process terms after slashes
    for term in terms[1:]:
        # Split units and append a -1 to the first one because now it is a unit
        # after a slash and append the rest without changes because they are
        # preceeded by dots.
        sub_terms = term.split('.')
        result.append(f"{sub_terms[0]}-1")
        result.extend(sub_terms[1:])
    # Join all terms with a space
    return ' '.join(result)



def _parse_madcuba_friendly_bunit(unit):
    """
    This function returns a BUNIT card string that MADCUBA can recognize.
    """
    if not isinstance(unit, u.UnitBase):
        raise TypeError("Input value is not an Astropy unit")
    else:
        if unit == u.Jy * u.m / u.beam / u.s:
            return 'Jy beam-1 m s-1'
        elif unit == u.Jy * u.km / u.beam / u.s:
            return 'Jy beam-1 km s-1'
        elif unit == u.mJy * u.m / u.beam / u.s:
            return 'mJy beam-1 m s-1'
        elif unit == u.mJy * u.km / u.beam / u.s:
            return 'mJy beam-1 km s-1'
        elif unit == u.Jy / u.beam:
            return 'Jy beam-1'
        elif unit == u.mJy / u.beam:
            return 'mJy beam-1'
        else:
            return unit.to_string(fraction=False)



import numpy as np

__all__ = [
    'transform_coords_fitsmap',
    'transform_coords_axes',
]

def transform_coords_fitsmap(
        ref_fitsmap,
        target_fitsmap,
        points,
        origin=0):
    """Transform the pixel coordinates of one or more points between
    two images.

    Parameters
    ----------
    ref_fitsmap : `~astropy.nddata.CCDData`
        Reference map.
    target_fitsmap : `~astropy.nddata.CCDData`
        Map into which the points are transformed.
    points : `~numpy.ndarray`
        Points from ref_fitsmap to transform.
    origin : `int`, default:0
        Origin of the coordinates of the image. 0 for numpy standards, and 1
        for FITS standards.

    Returns
    -------
    new_points : `~numpy.ndarray`
        Transformed points in the second image. 

    """

    if isinstance(points, tuple):
        points = np.array(points)

    if ref_fitsmap.header['NAXIS'] == 2:
        # Pixels to coordinates
        ra, dec = \
            ref_fitsmap.wcs.wcs_pix2world(points.T[0], points.T[1], origin)
        # Coordinates to pixels from second image
        new_x, new_y = \
            target_fitsmap.wcs.wcs_world2pix(ra, dec, origin)
    elif ref_fitsmap.header['NAXIS'] == 4:
        # Pixels to coordinates
        ra, dec, frec_world, pol_world = \
            ref_fitsmap.wcs.wcs_pix2world(
                points.T[0], points.T[1], 1, 1, origin
            )
        # Coordinates to pixels from second image
        new_x, new_y, nul_frec_pix, nul_pol_pix = \
            target_fitsmap.wcs.wcs_world2pix(
                ra, dec, frec_world, pol_world, origin
            )
    
    new_points = np.array([new_x, new_y]).T

    return new_points


def transform_coords_axes(
        ref_ax,
        target_ax,
        points):
    """Transform the pixel coordinates of one or more points between
    two WCSAxes.

    Parameters
    ----------
    ref_ax : `~astropy.visualization.wcsaxes.WCSAxes`
        Reference axes.
    target_ax : `~astropy.visualization.wcsaxes.WCSAxes`
        Axes into which the points are transformed.
    points : `~numpy.ndarray`
        Points from ref_axes to transform.

    Returns
    -------
    new_points : `~numpy.ndarray`
        Transformed points in the second image. 

    """

    # Astropy does not provide a straightforward method to get world
    # coordinates from an image pixel. It does however provide
    # transformations between pixel or world coordinates and the screen
    # coordinates. To get the world coordinates using only the WCSAxes
    # object, we must convert from pixels to screen, and then screen to
    # pixels.
    reference_pix2screen = ref_ax.get_transform('pixel')
    reference_screen2world = ref_ax.get_transform('world').inverted()
    reference_screen_coords = reference_pix2screen.transform(points)
    world_coords = reference_screen2world.transform(reference_screen_coords)
    # Finally we can do the same but backwards for the second axes.
    final_world2screen = target_ax.get_transform('world')
    final_screen2pix = target_ax.get_transform('pixel').inverted()
    final_screen_coords = final_world2screen.transform(world_coords)
    new_points = final_screen2pix.transform(final_screen_coords)

    return new_points



# madcubapy/visualization/__init__.py

"""
The `madcubapy.visualization` package provides functions to help the user
visualize astronomical data, with a special emphasis on products exported by
MADCUBA.
"""

from .interaction import *
from .quick_plotters import *
from .wcsaxes_helpers import *
from .zoom_syncers import *



from astropy.table import Table
import astropy.units as u
from astropy.utils.diff import report_diff_values
import numpy as np
import pytest
from madcubapy.io.madcubamap import MadcubaMap

@pytest.fixture
def example_madcuba_map():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "examples/data/IRAS16293_SO_2-1_moment0_madcuba.fits"
    )

@pytest.fixture
def example_carta_map():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "examples/data/IRAS16293_SO2c_moment0_carta.fits"
    )

def test_read_madcuba_map(example_madcuba_map):
    # Test if the obligatory attributes are correctly initialized
    assert example_madcuba_map.ccddata is not None
    assert example_madcuba_map.data is not None
    assert example_madcuba_map.header is not None
    assert example_madcuba_map.filename is not None
    assert (example_madcuba_map.hist is None or
            isinstance(example_madcuba_map.hist, Table))
    assert np.array_equal(
        example_madcuba_map.data,
        example_madcuba_map.ccddata.data,
        equal_nan=True,
    )
    assert (example_madcuba_map.hist[-1]["Macro"] ==
        "//PYTHON: Open cube: 'examples/data/IRAS16293_SO_2-1_moment0_madcuba.fits'"
    )

def test_invalid_file():
    with pytest.raises(FileNotFoundError):
        MadcubaMap.read("nonexistent_file.fits")

def test_write_madcuba_map(example_madcuba_map):
    # Assert filename and hist change after writing
    assert example_madcuba_map.filename == "IRAS16293_SO_2-1_moment0_madcuba.fits"
    example_madcuba_map.write(
        "examples/data/IRAS16293_SO_2-1_moment0_madcuba_write.fits",
        overwrite=True,
    )
    assert example_madcuba_map.filename == "IRAS16293_SO_2-1_moment0_madcuba_write.fits"
    assert (example_madcuba_map.hist[-1]["Macro"] ==
        "//PYTHON: Save cube: 'examples/data/IRAS16293_SO_2-1_moment0_madcuba_write.fits'"
    )
    # Read back the written file
    example_madcuba_map_write = MadcubaMap.read(
        "examples/data/IRAS16293_SO_2-1_moment0_madcuba_write.fits"
    )
    # Assert BUNIT was written correctly
    assert example_madcuba_map_write.header["BUNIT"] == 'Jy beam-1 m s-1'
    # Assert written map is correctly read
    assert 0 == 0
    assert example_madcuba_map_write.ccddata is not None
    assert example_madcuba_map_write.data is not None
    assert example_madcuba_map_write.header is not None
    assert example_madcuba_map_write.filename is not None
    assert (example_madcuba_map_write.hist is None or
            isinstance(example_madcuba_map_write.hist, Table))
    # Assert written map is equal to original map
    assert np.array_equal(
        example_madcuba_map.data,
        example_madcuba_map_write.data,
        equal_nan=True,
    )
    assert example_madcuba_map.unit == example_madcuba_map_write.unit
    assert (example_madcuba_map.hist[0]["Macro"] ==
        example_madcuba_map_write.hist[0]["Macro"]
    )
    # Check filename and hist file
    assert example_madcuba_map_write.filename == "IRAS16293_SO_2-1_moment0_madcuba_write.fits"
    assert (example_madcuba_map_write.hist[-2]["Macro"] ==
        "//PYTHON: Save cube: 'examples/data/IRAS16293_SO_2-1_moment0_madcuba_write.fits'"
    )
    assert (example_madcuba_map_write.hist[-1]["Macro"] ==
        "//PYTHON: Open cube: 'examples/data/IRAS16293_SO_2-1_moment0_madcuba_write.fits'"
    )

def test_fix_units_correct(example_madcuba_map):
    assert example_madcuba_map.unit == u.Jy * u.m / u.beam / u.s
    example_madcuba_map.fix_units()
    assert example_madcuba_map.unit == u.Jy * u.m / u.beam / u.s

def test_fix_units_incorrect(example_carta_map):
    assert example_carta_map.unit == u.Jy / u.beam / u.km / u.s
    example_carta_map.fix_units()
    assert example_carta_map.unit == u.Jy * u.km / u.beam / u.s

def test_copy_madcuba(example_madcuba_map):
    madcuba_map_copy = example_madcuba_map.copy()
    assert np.array_equal(
        madcuba_map_copy.data, example_madcuba_map.data, equal_nan=True
    )
    assert madcuba_map_copy.header == example_madcuba_map.header
    assert madcuba_map_copy.unit == example_madcuba_map.unit
    assert report_diff_values(example_madcuba_map.hist, madcuba_map_copy.hist)
    assert madcuba_map_copy.ccddata.meta == example_madcuba_map.ccddata.meta

def test_convert_units_madcuba(example_madcuba_map):
    example_madcuba_map_mJy = example_madcuba_map.copy()
    example_madcuba_map_mJy.convert_unit_to(u.mJy * u.m / u.beam / u.s)
    assert example_madcuba_map_mJy.unit == u.mJy * u.m / u.beam / u.s
    assert example_madcuba_map_mJy.ccddata.unit == u.mJy * u.m / u.beam / u.s
    assert example_madcuba_map_mJy.hist[-1]["Macro"] == (
        "//PYTHON: Convert units to 'm mJy beam-1 s-1'"
    )

def test_convert_units_carta(example_carta_map):
    example_carta_map_mJy = example_carta_map.copy()
    example_carta_map_mJy.fix_units()
    example_carta_map_mJy.convert_unit_to(u.mJy * u.m / u.beam / u.s)
    assert example_carta_map_mJy.unit == u.mJy * u.m / u.beam / u.s
    assert example_carta_map_mJy.ccddata.unit == u.mJy * u.m / u.beam / u.s
    assert example_carta_map_mJy.hist == None



import astropy.units as u
import numpy as np
import pytest
from madcubapy.io.madcubamap import MadcubaMap
from madcubapy.coordinates.offsets import get_angular_offset_points
from madcubapy.coordinates.offsets import get_physical_offset_points
from madcubapy.coordinates.offsets import get_angular_separation
from madcubapy.coordinates.offsets import get_physical_separation

@pytest.fixture
def example_madcuba_map():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "examples/data/IRAS16293_SO_2-1_moment0_madcuba.fits"
    )

def test_get_angular_offset_points(example_madcuba_map):
    # Test a specific array of points
    points = np.array([[120, 250], [140, 250], [160,250], [180,250]])
    point_B = (370, 300)
    points_angle = get_angular_offset_points(
        points, point_B, example_madcuba_map)
    assert np.allclose(
        points_angle,
        np.array([[0.01597222, -0.00319444],
                  [0.01469444, -0.00319444],
                  [0.01341667, -0.00319444],
                  [0.01213889, -0.00319444]]) * u.deg
    )

def test_get_physical_offset_points(example_madcuba_map):
    # Test a specific array of points
    points = np.array([[120, 250], [140, 250], [160,250], [180,250]])
    point_B = (370, 300)
    distance = 141 * u.pc
    points_physical = get_physical_offset_points(
        points, point_B, example_madcuba_map, distance)
    assert np.allclose(
        points_physical,
        np.array([[8107.50002507, -1621.50000501],
                  [7458.90002306, -1621.50000501],
                  [6810.30002106, -1621.50000501],
                  [6161.70001905, -1621.50000501]]) * u.au
    )

def test_get_angular_separation(example_madcuba_map):
    # Test a specific array of points
    points = np.array([[120, 250], [140, 250], [160,250], [180,250]])
    point_B = (370, 300)
    separation_angle = get_angular_separation(
        points, point_B, example_madcuba_map)
    assert np.allclose(
        separation_angle,
        np.array([0.01628853, 0.01503766, 0.01379172, 0.01255218]) * u.deg
    )

def test_get_physical_separation(example_madcuba_map):
    # Test a specific array of points
    points = np.array([[120, 250], [140, 250], [160,250], [180,250]])
    point_B = (370, 300)
    distance = 141 * u.pc
    separation_physical = get_physical_separation(
        points, point_B, example_madcuba_map, distance)
    assert np.allclose(
        separation_physical,
        np.array([8268.06016686,
                  7633.11547275,
                  7000.67487055,
                  6371.48408074]) * u.au
    )



import matplotlib.pyplot as plt
import numpy as np
import pytest
from madcubapy.coordinates import transform_coords_axes
from madcubapy.coordinates import transform_coords_fitsmap
from madcubapy.io.madcubamap import MadcubaMap
from madcubapy.visualization import add_wcs_axes

@pytest.fixture
def example_madcuba_map_m0():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "examples/data/IRAS16293_SO_2-1_moment0_madcuba.fits"
    )

@pytest.fixture
def example_madcuba_map_cont():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "examples/data/IRAS16293_band_7_tm2_cont.fits"
    )


def test_both_transforms(example_madcuba_map_m0, example_madcuba_map_cont):
    # Test if both functions yield the same points for the same plotted maps
    # Create fig
    fig = plt.figure(figsize=(3,3))
    ax1, img = add_wcs_axes(fig, 1, 2, 1, fitsmap=example_madcuba_map_m0)
    ax2, img = add_wcs_axes(fig, 1, 2, 2, fitsmap=example_madcuba_map_cont)
    # Original points
    points = np.array([[300, 350], [300, 400], [300, 450]])
    # Transform points from ccddata
    new_points_ccddata = transform_coords_fitsmap(
        ref_fitsmap=example_madcuba_map_m0,
        target_fitsmap=example_madcuba_map_cont,
        points=points
    )
    # Transform points from axes
    new_points_axes = transform_coords_axes(
        ref_ax=ax1,
        target_ax=ax2,
        points=points
    )
    assert np.allclose(new_points_ccddata, new_points_axes)



import astropy
import astropy.units as u
import numpy as np

__all__ = [
    'create_spectral_array',
]

def create_spectral_array(nchan, cdelt, crpix, crval):
    """
    Create a spectral axis array.

    Parameters
    ----------
    nchan : `~int`
        Number of channels in the spectrum.
    cdelt : `~float` or `~astropy.units.Quantity`
        Width of a channel.
    crpix : `~float`
        Reference channel of the spectrum.
    crval : `~float`
        Value of the reference channel.
    
    Returns
    -------
    spectral_array : `~numpy.ndarray` or `~astropy.units.Quantity`
        Returned spectral axis array with units if cdelt is a quantity.

    """

    # if cdelt has units
    if isinstance(cdelt, astropy.units.Quantity):
        first_chan = crval - (cdelt.value * crpix)
        last_chan = crval + (cdelt.value * (nchan-1 - crpix))
        spectral_array = np.linspace(first_chan, last_chan, nchan) * cdelt.unit
    # if cdelt is adimensional
    else:
        first_chan = crval - (cdelt * crpix)
        last_chan = crval + (cdelt * (nchan-1 - crpix))
        spectral_array = np.linspace(first_chan, last_chan, nchan)

    return spectral_array
    


import astropy.units as u
import numpy as np

__all__ = [
    'get_angular_offset_points',
    'get_angular_separation',
    'get_physical_offset_points',
    'get_physical_separation',
]

def get_angular_offset_points(points, ref_point, fitsmap):
    """Get offset coordinates of a point or set of points, with respect to a
    reference point in angular units.

    Parameters
    ----------
    points : `~numpy.ndarray`
        Input point or points.
    ref_point : `~numpy.ndarray`
        Reference point.
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map from which to get angular size of each pixel (CDELT1, CDELT2).

    Returns
    -------
    angular_offset_points : `~astropy.units.Quantity`
        Points with converted units to angular offset.

    """

    cdelt_deg = fitsmap.wcs.wcs.cdelt[:2] * u.deg
    angular_offset_points = (points - ref_point) * cdelt_deg

    return angular_offset_points


def get_angular_separation(points, ref_point, fitsmap):
    """Get separation between a point or set of points, and a reference pixel
    in angular units.

    Parameters
    ----------
    points : `~numpy.ndarray`
        Input point or points.
    ref_point : `~numpy.ndarray`
        Reference point.
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map from which to get angular size of each pixel (CDELT1, CDELT2).

    Returns
    -------
    angular_separation : `~astropy.units.Quantity`
        Separation between input points and reference point.

    """

    angular_offset_points = get_angular_offset_points(points, ref_point, fitsmap)
    angular_separation = np.linalg.norm(angular_offset_points, axis=-1)

    return angular_separation


def get_physical_offset_points(points, ref_point, fitsmap, distance):
    """Get offset coordinates of a point or set of points, with respect to a
    reference point in physical distance units.

    Parameters
    ----------
    points : `~numpy.ndarray`
        Input point or points.
    ref_point : `~numpy.ndarray`
        Reference point.
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map from which to get angular size of each pixel (CDELT1, CDELT2).
    distance : `~astropy.units.quantity`
        Distance from earth to object.

    Returns
    -------
    physical_offset_points : `~astropy.units.Quantity`
        Points with converted units to physical distance offset.

    """

    angular_offset_points = get_angular_offset_points(points, ref_point, fitsmap)

    physical_offset_points = (angular_offset_points * distance).to(
        u.au, equivalencies=u.dimensionless_angles()
    )

    return physical_offset_points


def get_physical_separation(points, ref_point, fitsmap, distance):
    """Get separation between a point or set of points, and a reference pixel
    in physical distance units.

    Parameters
    ----------
    points : `~numpy.ndarray`
        Input point or points.
    ref_point : `~numpy.ndarray`
        Reference point.
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map from which to get angular size of each pixel (CDELT1, CDELT2).
    distance : `~astropy.units.quantity`
        Distance from earth to object.

    Returns
    -------
    physical_separation : `~astropy.units.Quantity`
        Separation between input points and reference point.

    """

    physical_offset_points = get_physical_offset_points(points, ref_point, fitsmap, distance)
    physical_separation = np.linalg.norm(physical_offset_points, axis=-1)

    return physical_separation



import astropy.units as u
import numpy as np
import pytest
from madcubapy.utils.spectral import create_spectral_array

def test_create_spectral_array_without_units():
    # Test a specific array without units
    assert (create_spectral_array(8, 0.5, -3, 10).all()
         == np.array([11.5, 12, 12.5, 13, 13.5, 14, 14.5, 15]).all())

def test_create_spectral_array_with_units():
    # Test a specific array with units
    a = create_spectral_array(8, 0.5 * u.s, -3, 10)
    b = np.array([11.5, 12, 12.5, 13, 13.5, 14, 14.5, 15]) * u.s
    assert a.value.all() == b.value.all()
    assert a.unit == b.unit
